#!/bin/bash
# This file is an installation script for the W&T USB Redirector Software for Linux by Wiesemann & Theis GmbH.


# Binary path
DEST_BINNAME="usb-redirector"
DEST_BINDIR="/usr/local/bin/"

# Systemd path
DEST_INITDIR_SYSTEMD="/etc/systemd/system"
DEST_INITPOSTFIX_DAEMON="-daemon.service"
DEST_INIT_SYSTEMD_DAEMON=""

# Source path
BINNAME="usb-redirector"
SYSTEM_DAEMON="usb-redirector-daemon.service"

# Save current directory and change tocheckModprobeLoadVhci
cd $(dirname $0)

# Utils
echoerr() { echo "$@" 1>&2; }

echoOk() {
	echo -e "\e[32mOK\e[0m"
}

echoNo() {
	echo -e "\e[35mNo\e[0m"
}

echoFail() {
	echo -e "\e[31mFAIL\e[0m"
}

myExit() {
	# change back to user directory
	popd 2>> /dev/null
	exit $@
}

# Global flags/variables
systemdSupport=0
systemdEnable=0
kernelVersionMajor=0
kernelVersionMinor=0
systemArch=""

# Functions
function checkForFiles {
    echo -n "Checking whether $BINNAME.x86_64 file exists: "
    stat "$BINNAME.x86_64" &> /dev/null
	if [ $? -eq 0 ]; then
		echoOk
	else 
		echoFail
		myExit -1
	fi

	echo -n "Checking whether $BINNAME.i686 file exists: "
	stat "$BINNAME.i686" &> /dev/null
	if [ $? -eq 0 ]; then
		echoOk
	else 
		echoFail
		myExit -1
	fi

	echo -n "Checking whether $SYSTEM_DAEMON file exists: "
	stat "$SYSTEM_DAEMON" &> /dev/null
	if [ $? -eq 0 ]; then
		echoOk
	else 
		echoFail
		myExit -1
	fi
}

function checkKernel {
	echo -en "Checking kernel version: "
	version=$(uname -r | grep -oE '^[0-9]+.[0-9]+')
	majVersion=$(echo $version | grep -oE "^[0-9]+")
	minVersion=$(echo $version | grep -oE "[0-9]+$")

	if [ $majVersion -lt 4 ] || ( [ $majVersion -eq 4 ] && [ $minVersion -lt 4 ] ); then
		echoFail
		echoerr "Your kernel version $version is too old and not supported by this software. Please update to a kernel version >= 4.4!"
		myExit -1;
	else
		echoOk
		kernelVersionMajor=$majVersion
		kernelVersionMinor=$minVersion
	fi
}

function checkForSystemdSupport {
	echo -en "Checking for systemd support: "

	if [ -d "/run/systemd/system/" ]; then
		echoOk
        inp=""
        while [ "$inp" != "y" ] && [ "$inp" != "Y" ] && [ "$inp" != "n" ] && [ "$inp" != "N" ]; do
            read -p "Your system supports the systemd init system. Do you want to install systemd init scripts (WARNING: overwrites existing service files)? (yY/nN)" inp
            if [ "$inp" == "y" ] || [ "$inp" == "Y" ]; then
                systemdSupport=1;
                inp=""
                while [ "$inp" != "y" ] && [ "$inp" != "Y" ] && [ "$inp" != "n" ] && [ "$inp" != "N" ]; do
                    read -p "Do you want to enable the USB redirector daemon service on system start? (yY/nN)" inp
                    if [ "$inp" == "y" ] || [ "$inp" == "Y" ]; then
                        systemdEnable=1;
                    fi
                done
            elif [ "$inp" == "n" ] || [ "$inp" == "N" ]; then
                echo "User disabled systemd support manually."
            fi
        done
	else
		echoNo
	fi
}

# function checkForVhciSupport {
# 	echo -en "Checking for VHCI/USBIP support: "

# 	if [ -f "/lib/modules/$(uname -r)/kernel/drivers/usb/usbip/vhci-hcd.ko" ]; then
# 		echoOk
# 	else
# 		echoFail
# 		echoerr "The VHCI kernel module is required! Please install the usbip package."
# 		myExit -1
# 	fi
# }

function checkForModprobeBin {
	echo -en "Checking for modprobe executable: "

	if [ -n "$(which modprobe 2> /dev/null)" ] || [ -n "$(ls /sbin/ | grep modprobe)" ]; then
		echoOk
	else
		echoFail
		echoerr "The modprobe program is not installed! Please install modprobe."
		myExit -1
	fi
}

function checkForArchitectureSupport {
	echo -en "Checking for supported architecture: "

	arch="$(uname -m)"

	if [ "$arch" == "x86_64" ]; then
		echoOk
		systemArch="$arch"
	elif [ "$arch" == "i686" ]; then
		echoOk
		systemArch="$arch"
	else
		echoFail
		myExit -1
	fi
}

function checkForLinkedLib {
	echo -n "Checking for required libraries: "
	if [ -n "$systemArch" ]; then
		if [ -f "$BINNAME.$systemArch" ]; then
			libs=$(ldd $BINNAME.$systemArch)
			if [ $? -ne 0 ]; then
				echoFail
				echo "Could not check libs for $BINNAME.$systemArch binary with ldd!"
				myExit -1
			else
				ldd $BINNAME.$systemArch | while read e; do
					if [ -n "$(echo $e | grep 'not found')" ]; then
						lib="$(echo $e | grep -Eo '^[^ ]+')"
						echoFail
						echo "The $lib library is required but not installed on your system! Please install $lib."
						myExit -1
					fi
				done
			fi
		else
			echoFail
			echo "Could not find USB Redirector binary for $systemArch system architecture."
			myExit -1
		fi
	else
		echoFail
		echo "No valid system architecture detected."
		myExit -1
	fi
	echoOk
}

function checkModprobeLoadVhci {
	echo -n "Trying to (un)load vhci kernel module via modprobe: "
	modprobe vhci_hcd
	if [ $? -ne 0 ]; then
		echoFail
		echo "Could not load vhci kernel module via 'modprobe vhci_hcd'."
        if [ -f "/lib/modules/$(uname -r)/kernel/drivers/usb/usbip/vhci-hcd.ko" ] || [ -f "/lib/modules/$(uname -r)/kernel/drivers/usb/usbip/vhci-hcd.ko.zst" ]; then
            echo "It seems that the Kernelmodule vhci-hcd is correct installed..." 
        else
            echo "The Kernelmodule vhci-hcd could be not found in /lib/modules/$(uname -r)/kernel/drivers/usb/usbip/" 
        fi
		myExit -1
	fi

	modprobe -r vhci_hcd
	if [ $? -ne 0 ]; then
		echoFail
		echo "Could not unload vhci kernel module via 'modprobe -r vhci_hcd'."
		myExit -1
	fi
	echoOk
}

function checkForRoot {
	if [ "$(id -nu)" != "root" ]; then
		echo ""
		echo "Installing 'W&T USB Redirector Software for Linux' requires administrator privileges. Please run this script as root user:"
		echo "sudo $0"
		echo ""
		myExit 1
	fi
}

function getInstallDirectoriesAndPath {
	echo "Next you can specify filenames and installation paths. To use the default settings just press <enter> without making any entry."
	echo ""
	sleep 1

	read -p "Please enter the filename for the binary [$DEST_BINNAME]: " newPath
	DEST_BINNAME="${newPath:-$DEST_BINNAME}"

	read -p "Please enter the installation path for the binary [$DEST_BINDIR]: " newPath
	DEST_BINDIR="${newPath:-$DEST_BINDIR}"

	if [ $systemdSupport -eq 1 ]; then
		read -p "Please enter the installation path for the systemd service [$DEST_INITDIR_SYSTEMD]: " newPath
		DEST_INITDIR_SYSTEMD="${newPath:-$DEST_INITDIR_SYSTEMD}"
		read -p "Please enter the service file name for the systemd service [$DEST_BINNAME$DEST_INITPOSTFIX_DAEMON]: " newPath
		DEST_INIT_SYSTEMD_DAEMON="${newPath:-$DEST_BINNAME$DEST_INITPOSTFIX_DAEMON}"
	fi

	echo ""
	echo -e "binary name:\t\t\t\t$DEST_BINNAME"
	echo -e "binary path:\t\t\t\t$DEST_BINDIR"
	if [ $systemdSupport -eq 1 ]; then
		echo -e "systemd service path:\t\t\t$DEST_INITDIR_SYSTEMD"
		echo -e "systemd service name for daemon:\t$DEST_INIT_SYSTEMD_DAEMON"
	fi
	echo ""
    inp=""
    while [ "$inp" != "y" ] && [ "$inp" != "Y" ] && [ "$inp" != "n" ] && [ "$inp" != "N" ]; do
        read -p "Do you want to continue with the installation? (yY/nN) " inp 
        if [ ! "$inp" == "y" ] && [ ! "$inp" == "Y" ]; then
            echo ""
            echo "Installation aborted! Please note that nothing has been changed or installed."
            echo ""
            myExit 0;
        fi
    done
}

function installBinary {
	echo -n "Creating binary directory [$DEST_BINDIR]: "
	mkdir -p $DEST_BINDIR >> /dev/null
	if [ $? -ne 0 ] || [ ! -d "$DEST_BINDIR" ]; then
		echoFail
		echo "Could not create $DEST_BINDIR directory!"
		myExit -1
	else
		echoOk
	fi

	echo -n "Installing binary to directory [$BINNAME.$systemArch -> $DEST_BINDIR/$DEST_BINNAME]: "
	rm -f "$DEST_BINDIR/$DEST_BINNAME" >> /dev/null
	cp -af "$BINNAME.$systemArch" "$DEST_BINDIR/$DEST_BINNAME" >> /dev/null
	if [ $? -ne 0 ]; then
		echoFail
		echo "Could not copy $BINNAME.$systemArch binary to $DEST_BINDIR/$DEST_BINNAME directory!"
		myExit -1
	else
		echoOk
	fi

	echo -n "Changing permission to 711: "
	chmod 711 $DEST_BINDIR/$DEST_BINNAME >> /dev/null
	if [ $? -ne 0 ]; then
		echoFail
		echo "Could not change permission to 711 for $DEST_BINDIR/$DEST_BINNAME binary!"
		myExit -1
	else
		echoOk
	fi
}
function systemdReload() {
	echo -n "Reloading systemd: "
	systemctl daemon-reload >> /dev/null
	if [ $? -ne 0 ]; then
		echoNo
		echo "Could not reload systemd!"
	else
		echoOk
	fi
}

function systemdEnableDaemon() {
	echo -n "Enabling $DEST_INITDIR_SYSTEMD/$DEST_INIT_SYSTEMD_DAEMON service: "
	if [ ! "$(systemctl is-enabled $DEST_INIT_SYSTEMD_DAEMON)" == "enabled" ]; then
		systemctl enable $DEST_INIT_SYSTEMD_DAEMON >> /dev/null
		if [ $? -ne 0 ]; then
			echoNo
			echo "Could not enable $DEST_INIT_SYSTEMD_DAEMON system service!"
		else
			echoOk
		fi
	else
		echoOk
	fi
	systemdStartDaemon
}

function systemdStartDaemon() {
	echo -n "Start $DEST_INITDIR_SYSTEMD/$DEST_INIT_SYSTEMD_DAEMON service: "
	systemctl start $DEST_INIT_SYSTEMD_DAEMON >> /dev/null
	if [ ! "$(systemctl is-active $DEST_INIT_SYSTEMD_DAEMON)" == "active" ]; then
		echoNo
		echo "Could not start system service $DEST_INIT_SYSTEMD_DAEMON"
	else
		echoOk
	fi
}

function installStartscriptSystemd() {
	echo -n "Installing daemon service file to $DEST_INITDIR_SYSTEMD/$DEST_INIT_SYSTEMD_DAEMON: "
	mkdir -p $DEST_INITDIR_SYSTEMD >> /dev/null
	if [ $? -ne 0 ] || [ ! -d "$DEST_INITDIR_SYSTEMD" ]; then
		echoFail
		echo "Could not create $DEST_INITDIR_SYSTEMD directory!"
		myExit -1
	else
		rm -f "$DEST_INITDIR_SYSTEMD/$DEST_INIT_SYSTEMD_DAEMON" >> /dev/null
		cp -af $SYSTEM_DAEMON $DEST_INITDIR_SYSTEMD/$DEST_INIT_SYSTEMD_DAEMON >> /dev/null
		if [ $? -ne 0 ] || [ ! -f "$DEST_INITDIR_SYSTEMD/$DEST_INIT_SYSTEMD_DAEMON" ]; then
			echoFail
			echo "Could not install $SYSTEM_DAEMON to $DEST_INITDIR_SYSTEMD/$DEST_INIT_SYSTEMD_DAEMON!"
			myExit -1
		else
			echoOk
			systemdReload
			if [ $systemdEnable -eq 1 ]; then
				systemdEnableDaemon
			fi
		fi
	fi
}

function installStartscript() {
	if [ $systemdSupport -eq 1 ]; then
		installStartscriptSystemd
	fi
}

function checkForRunning {
	prog="$DEST_BINDIR/$DEST_BINNAME"
	serv="$DEST_BINNAME$DEST_INITPOSTFIX_DAEMON"
	if [ -n "$1" ]; then
		prog="$1"
	fi
	if [ -n "$2" ]; then
		serv="$2"
	fi

	if [ -n "$serv" ]; then
		if [ "$(systemctl is-active $serv)" == "active" ]; then
			echo "The systemd service $serv is currently running but must be stopped for this installation. If you have plugged any devices it is recommended to abort this installation and unplug all plugged devices first."
			inp=""
            while [ "$inp" != "y" ] && [ "$inp" != "Y" ] && [ "$inp" != "n" ] && [ "$inp" != "N" ]; do
                read -p "Do you want to continue with the installation and stop $serv? (yY/nN): " inp 
                if [ "$inp" == "y" ] || [ "$inp" == "Y" ]; then
                    if [ -f "$prog" ]; then
                        $prog unplug all &> /dev/null
                    fi
                    systemctl stop $serv
                    sleep 1
                fi
                if [ "$inp" == "n" ] || [ "$inp" == "N" ]; then
                    echo "Installation terminated by user"
                    exit 1
                fi
            done
		fi
	fi

	if [ -f "$prog" ]; then
		pidof $prog
		if [ $? -eq 0 ]; then
			echo "'$prog' is already running but must be stopped for this installation. If you have plugged any devices it is recommended to abort this installation, unplug all plugged devices and stop all instances of this binary first."
			inp=""
            while [ "$inp" != "y" ] && [ "$inp" != "Y" ] && [ "$inp" != "n" ] && [ "$inp" != "N" ]; do
                read -p "Do you want to continue with the installation and force a stop of all instances of the $prog binary? (yY/nN): " inp 
                if [ "$inp" == "y" ] || [ "$inp" == "Y" ]; then
                    cnt=0
                    $prog unplug all &> /dev/null
                    while [ $cnt -lt 10 ] && [ -n "$(pidof -s $prog)" ]; do
                        echo "Stopping PID $(pidof -s $prog)"
                        kill -9 "$(pidof -s $prog)"
                        sleep 1
                        cnt=$(( $cnt+1 ))
                    done
                fi
                if [ "$inp" == "n" ] || [ "$inp" == "N" ]; then
                    echo "Installation terminated by user"
                    exit 1
                fi
            done
		fi
	fi

}

##################################################################################################################################
echo ""
echo "###################################################################"
echo "#       USB Umlenkung für Linux / USB Redirector for Linux        #"
echo "#                     Wiesemann & Theis GmbH                      #"
echo "###################################################################"
echo ""

checkForRoot
checkForRunning

echo ""
echo "##### 1. Checking system requirements #####"
checkForFiles
checkKernel
checkForSystemdSupport
checkForModprobeBin
checkForArchitectureSupport
checkForLinkedLib
checkModprobeLoadVhci
echo ""
echo "Checking system requirements successfully completed."
echo ""

echo ""
echo "##### 2. Checking installation files and directories #####"
getInstallDirectoriesAndPath
checkForRunning
echo ""
echo "Checking installation files and directories successfully completed."
echo ""

echo ""
echo "##### 3. Installing files and directories #####"
installBinary
installStartscript
echo ""
echo "Installing files and directories successfully completed."
echo ""

echo ""
echo ""

printf '#%.0s' $(seq 1 $COLUMNS)
echo ""
echo ""
echo "The USB Redirector software was successessfully installed on your system."
echo "Use '$DEST_BINNAME -h' for general help (all commands) or '$DEST_BINNAME <command> -h' for specific commands (e. g. '$DEST_BINNAME scan -h')."
echo "Please also refer to the instruction manual ('manual.pdf')."
if [ $systemdSupport -eq 1 ]; then
	echo "Use 'systemctl status $DEST_INIT_SYSTEMD_DAEMON' to get the current daemon status."
	echo "If you want to install a start script to plug an usb device on system start, you can use the template service file 'usb-redirector-plug-devices.service' and copy the file to '$DEST_INITDIR_SYSTEMD'."
	echo "Remember to call 'systemctl daemon-reload' before you can use the newly installed service via 'systemctl [enable|start|status|restart] usb-redirector-plug-devices.service'."
fi
echo ""
echo "If you have any further questions, don't hesitate to contact the Wiesemann & Theis GmbH support."
echo ""
printf '#%.0s' $(seq 1 $COLUMNS)
echo ""
echo ""

myExit 0